@@ -1,4 +1,4 @@
-from kadmin_local._lib import exceptions
+from kadmin._lib import exceptions
 
 PyKAdminException = exceptions.PyKAdminException
 KAdminException = exceptions.KAdminException
@@ -7,6 +7,9 @@ NullPointerDereference = exceptions.NullPointerDereference
 CStringConversion = exceptions.CStringConversion
 CStringImportFromVec = exceptions.CStringImportFromVec
 StringConversion = exceptions.StringConversion
+ThreadSendError = exceptions.ThreadSendError
+ThreadRecvError = exceptions.ThreadRecvError
+TimestampConversion = exceptions.TimestampConversion
 
 __all__ = (
     "PyKAdminException",
@@ -16,4 +19,7 @@ __all__ = (
     "CStringConversion",
     "CStringImportFromVec",
     "StringConversion",
+    "ThreadSendError",
+    "ThreadRecvError",
+    "TimestampConversion",
 )
